#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2014-2017 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

"""
It provides sphinx extensions.

Extensions:

.. currentmodule:: schedula.ext

.. autosummary::
    :nosignatures:
    :toctree: exl/

    autosummary
    dispatcher
"""